require 'gitroom/models/node'
require 'yaml'

class Integer
  def minutes
    self * 60
  end
end

module NodeCached
  CACHE_FILES = 16384
  MAX_FILE_SIZE = 65536
  EXPIRE_TIME = 1.minutes
  READ_SCORE = 2
  WRITE_SCORE = -7

  # redis.expire key, seconds

  def find_cached_node(path)
    present = cached_node_present? path
    if present.nil? || present == 'true'
      node = Node.where(:path => path).first
      if node && present.nil?
        cached_node_present! path
      elsif !node
        cached_node_absent! path
      end
      node
    end
  end

  def read_cached_node(node, size, offset)
    path = node.path
    cached_file_rank_change! path, READ_SCORE
    rank = cached_file_rank path
    content = cached_file_content path
    if content
      print "CACHE HIT :: #{path}\n"
      res = content[offset...offset + size]
      if rank < CACHE_FILES
        cached_file_content_touch! path, EXPIRE_TIME
      end
    else
      print "DATABASE HIT :: #{path}\n"
      content = node.get_full_content
      full_size = content.size
      if rank < CACHE_FILES && full_size < MAX_FILE_SIZE
        res = content[offset...offset + size]
        cache_file! path, content
        cached_file_content_touch! path, EXPIRE_TIME
      else
        res = node.get_content offset, size
      end
    end
    res
  end

  def write_cached_node(node, buf, offset)
    path = node.path
    cached_file_rank_change! path, WRITE_SCORE
    cached_file_clean! path
    node.set_content offset, buf
    # read_cached_node node, buf.size, offset
  end

  def cached_node_present!(path)
    redis.set "cache::node::#{path}::present?", 'true'
  end

  def cached_node_absent!(path)
    redis.set "cache::node::#{path}::present?", 'false'
    redis.del "cache::file::#{path}::content"
    redis.zrem "cache::files::ranks", path
  end

  def cached_node_present?(path)
    redis.get "cache::node::#{path}::present?"
  end

  def cached_file_rank_change!(path, incr)
    redis.zincrby "cache::files::ranks", incr, path
  end

  def cached_file_rank(path)
    redis.zrevrank "cache::files::ranks", path
  end

  def cached_file_content(path)
    content = redis.get "cache::file::#{path}::content"
    content && YAML.load(content)
  end

  def cache_file!(path, content)
    redis.set "cache::file::#{path}::content", content.to_yaml
  end

  def cached_file_content_expires(path)
    redis.ttl "cache::file::#{path}::content"
  end

  def cached_file_content_touch!(path, time)
    redis.expire "cache::file::#{path}::content", time
  end

  def cached_file_clean!(path)
    redis.del "cache::file::#{path}::content"
  end
end
