require 'active_record'
require 'json'
require 'gitroom/models/chunk'

class Node < ActiveRecord::Base
  CHUNK_SIZE = 16384

  has_many :nodes, :dependent => :destroy
  belongs_to :node

  has_many :chunks, :dependent => :destroy

  def stat
    if directory
      RFuse::Stat.directory mode,:uid => uid, :gid => gid, :atime => updated_at,
                            :mtime => updated_at, :size => 48
    else
      RFuse::Stat.file mode,:uid => uid, :gid => gid, :atime => updated_at,
                       :mtime => updated_at, :size => content_size
    end
  end

  def getxattr(name)
    JSON.parse(xattrs)[name]
  end

  def setxattr(name, value, flag)
    json = JSON.parse(xattrs)
    json[name] = value
    xattrs = json.to_s
  end

  def removexattr(name)
    json = JSON.parse(xattrs)
    json.delete name
    xattrs = json.to_s
  end

  def listxattr
    JSON.parse(xattrs).keys
  end

  def get_chunk(number, offset = 0, right = CHUNK_SIZE - 1)
    chunk = chunks.where(:chunk_number => number).first
    chunk && chunk.content[offset..right] || ''
  end

  def get_chunk_cached(chunk, offset = 0, right = CHUNK_SIZE - 1)
    chunk && chunk.content[offset..right] || ''
  end

  def set_chunk(number, buf, offset = 0)
    chunk = chunks.where(:chunk_number => number).first
    unless chunk
      chunk = chunks.create! :chunk_number => number, :node_id => id, :content => ""
    end
    if chunk.content.size < offset+buf.size
      chunk.content += "\0" * (offset + buf.size - chunk.content.size)
    end
    chunk.content[offset...offset+buf.size] = buf
    rechunk = Chunk.find chunk.id
    rechunk.content = chunk.content.to_s
    rechunk.save!
  end

  def get_content(offset, size)
    num_start = offset / CHUNK_SIZE
    offset_start = offset % CHUNK_SIZE
    num_end = (offset + size - 1) / CHUNK_SIZE
    offset_end = (offset + size - 1) % CHUNK_SIZE
    content = ''
    cached_chunks = chunks.where('chunk_number >= ? and chunk_number <= ?', num_start, num_end).order('chunk_number asc')
    # (num_start..num_end).to_a.each do |num|
    cached_chunks.each do |chunk|
      num = chunk.chunk_number
      offset = num == num_start ? offset_start : 0
      right = num == num_end ? offset_end : (CHUNK_SIZE - 1)
      content += get_chunk_cached chunk, offset, right
    end
    content
  end

  def get_full_content
    chunks.order('chunk_number asc').pluck(:content).join
  end

  def set_content(offset, buf)
    size = buf.size
    num_start = offset / CHUNK_SIZE
    offset_start = offset % CHUNK_SIZE
    num_end = (offset + size - 1) / CHUNK_SIZE
    offset_end = (offset + size - 1) % CHUNK_SIZE
    (num_start..num_end).to_a.each do |num|
      offset = num == num_start ? offset_start : 0
      data_offset = num == num_start ? 0 : ((num - num_start) * CHUNK_SIZE - offset_start)
      data_right = num == num_end ? -1 : (data_offset + CHUNK_SIZE - 1)
      set_chunk num, buf[data_offset..data_right], offset
    end
    buf.size
  end

  # def get_content(offset, size)
  #   print 'A'
  #   num_start = offset / CHUNK_SIZE
  #   offset_start = offset % CHUNK_SIZE
  #   num_end = (offset + size - 1) / CHUNK_SIZE
  #   offset_end = (offset + size - 1) % CHUNK_SIZE
  #   print 'B'
  #   selected_chunks = chunks.where('chunk_number >= ? and chunk_number <= ?', num_start, num_end)
  #   selected_chunks = selected_chunks.pluck(:content)
  #   print 'C'
  #   if num_start == num_end || selected_chunks.size == 1
  #     print 'D0'
  #     selected_chunks[0] = selected_chunks[offset_start..offset_end]
  #   else
  #     print 'D1'
  #     selected_chunks[0] = selected_chunks[offset_start..-1]
  #     selected_chunks[-1] = selected_chunks[0..offset_end] if selected_chunks.count == num_end - num_start + 1
  #   end
  #   print 'E'
  #   res = selected_chunks.join
  #   print "F: #{res}"
  #   res
  # end

  # def set_content(offset, buf)
  #   print "\nset_content #{offset}, #{buf.size}:: \n#{buf}\n"
  #   Chunk.transaction do
  #     size = buf.size
  #     num_start = offset / CHUNK_SIZE
  #     offset_start = offset % CHUNK_SIZE
  #     num_end = (offset + size - 1) / CHUNK_SIZE
  #     offset_end = (offset + size - 1) % CHUNK_SIZE
  #     chunk_count = chunks.count
  #     last_chunk = chunks.last
  #     if chunk_count <= num_end
  #       if last_chunk && last_chunk.content.size < CHUNK_SIZE
  #         last_chunk.content += "\0" * (CHUNK_SIZE - last_chunk.content.size)
  #         rechunk = Chunk.find last_chunk.id
  #         rechunk.content = last_chunk.content.to_s
  #         rechunk.save!
  #       end
  #       while chunk_count <= num_end
  #         content = "\0" * (chunk_count == num_end ? offset_end + 1 : CHUNK_SIZE)
  #         chunks.create! :chunk_number => chunk_count, :node_id => id, :content => content
  #         chunk_count += 1
  #       end
  #     end
  #     selected_chunks = chunks.order('chunk_number asc').where('chunk_number >= ? and chunk_number <= ?', num_start, num_end)
  #     if num_start == num_end
  #       if selected_chunks[0].content.size <= offset_end
  #         selected_chunks[0].content += "\0" * (offset_end + 1 - selected_chunks[0].content.size)
  #       end
  #       selected_chunks[0].content[offset_start..offset_end] = buf
  #     else
  #       selected_chunks[0].content[offset_start..-1] = buf[0..CHUNK_SIZE - offset_start - 1]
  #       selected_chunks[-1].content[0..offset_end] = buf[-1 - offset_end..-1]

  #       selected_chunks[1...-1].each do |chunk|
  #         num = chunk.chunk_number - num_start
  #         chunk.content = buf[num*CHUNK_SIZE...(num+1)*CHUNK_SIZE]
  #       end
  #     end
  #     print "\n"
  #     selected_chunks.each do |chunk|
  #       rechunk = Chunk.find chunk.id
  #       rechunk.content = chunk.content.to_s
  #       print "set_content #{chunk.chunk_number} :: #{chunk.content}\n"
  #       rechunk.save!
  #     end
  #     print "\n"
  #   end
  # end

  def content_size
    if chunks.count == 0
      0
    else
      (chunks.count - 1) * CHUNK_SIZE + chunks.order('chunk_number asc').last.content.size
    end
  end

  def truncate_content(offset)
    Chunk.transaction do
      num_end = (offset) / CHUNK_SIZE
      offset_end = (offset) % CHUNK_SIZE
      selected_chunks = chunks.where('chunk_number > ?', num_end)
      selected_chunks.each { |chunk| chunk.destroy }
      last_chunk = chunks.last
      last_chunk.content = last_chunk.content[0..offset_end]
      last_chunk.save
    end
  end
end
